@@ -1,17 +1,29 @@
 # -*- coding: utf-8 -*-
 """Implements file-like objects for reading and writing from/to S3."""
-import boto3
-import botocore.client
 
 import io
+import contextlib
+import functools
+import itertools
 import logging
 
+import boto3
+import botocore.client
 import six
 
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
+# Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
+# The only method currently relying on it is s3_iter_bucket, which is instructed
+# whether to use it by the MULTIPROCESSING flag.
+_MULTIPROCESSING = False
+try:
+    import multiprocessing.pool
+    _MULTIPROCESSING = True
+except ImportError:
+    logger.warning("multiprocessing could not be imported and won't be used")
 
 START = 0
 CURRENT = 1
@@ -22,15 +34,13 @@ DEFAULT_MIN_PART_SIZE = 50 * 1024**2
 """Default minimum part size for S3 multipart uploads"""
 MIN_MIN_PART_SIZE = 5 * 1024 ** 2
 """The absolute minimum permitted by Amazon."""
-READ = 'r'
 READ_BINARY = 'rb'
-WRITE = 'w'
 WRITE_BINARY = 'wb'
-MODES = (READ, READ_BINARY, WRITE, WRITE_BINARY)
+MODES = (READ_BINARY, WRITE_BINARY)
 """Allowed I/O modes for working with S3."""
 
 BINARY_NEWLINE = b'\n'
-DEFAULT_BUFFER_SIZE = 256 * 1024
+DEFAULT_BUFFER_SIZE = 128 * 1024
 
 
 def _range_string(start, stop=None):
@@ -57,20 +67,14 @@ def open(bucket_id, key_id, mode, **kwargs):
     line_buffering = kwargs.pop("line_buffering", False)
     s3_min_part_size = kwargs.pop("s3_min_part_size", DEFAULT_MIN_PART_SIZE)
 
-    if mode in (READ, READ_BINARY):
+    if mode == READ_BINARY:
         fileobj = SeekableBufferedInputBase(bucket_id, key_id, **kwargs)
-    elif mode in (WRITE, WRITE_BINARY):
+    elif mode == WRITE_BINARY:
         fileobj = BufferedOutputBase(bucket_id, key_id, min_part_size=s3_min_part_size, **kwargs)
     else:
-        assert False
+        assert False, 'unexpected mode: %r' % mode
 
-    if mode in (READ, WRITE):
-        return io.TextIOWrapper(fileobj, encoding=encoding, errors=errors,
-                                newline=newline, line_buffering=line_buffering)
-    elif mode in (READ_BINARY, WRITE_BINARY):
-        return fileobj
-    else:
-        assert False
+    return fileobj
 
 
 class RawReader(object):
@@ -87,32 +91,58 @@ class RawReader(object):
 
 
 class SeekableRawReader(object):
-    """Read an S3 object.
+    """Read an S3 object."""
 
-    Support seeking around, but is slower than RawReader."""
     def __init__(self, s3_object):
-        self.position = 0
         self._object = s3_object
         self._content_length = self._object.content_length
+        self.seek(0)
+
+    def seek(self, position):
+        """Seek to the specified position (byte offset) in the S3 key.
+
+        :param int position: The byte offset from the beginning of the key.
+        """
+        self._position = position
+        range_string = _range_string(self._position)
+        logger.debug('content_length: %r range_string: %r', self._content_length, range_string)
+
+        #
+        # Close old body explicitly.
+        # When first seek(), self._body is not exist. Catch the exception and do nothing.
+        #
+        try:
+            self._body.close()
+        except AttributeError as e:
+            pass
+
+        if position == self._content_length == 0 or position == self._content_length:
+            #
+            # When reading, we can't seek to the first byte of an empty file.
+            # Similarly, we can't seek past the last byte.  Do nothing here.
+            #
+            self._body = io.BytesIO()
+        else:
+            self._body = self._object.get(Range=range_string)['Body']
 
     def read(self, size=-1):
-        if self.position == self._content_length:
+        if self._position >= self._content_length:
             return b''
-        if size <= 0:
-            end = None
+        if size == -1:
+            binary = self._body.read()
         else:
-            end = min(self._content_length, self.position + size)
-        range_string = _range_string(self.position, stop=end)
-        logger.debug('range_string: %r', range_string)
-        body = self._object.get(Range=range_string)['Body'].read()
-        self.position += len(body)
-        return body
+            binary = self._body.read(size)
+        self._position += len(binary)
+        return binary
 
 
 class BufferedInputBase(io.BufferedIOBase):
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
                  line_terminator=BINARY_NEWLINE, **kwargs):
-        session = boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        session = kwargs.pop(
+            's3_session',
+            boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        )
         s3 = session.resource('s3', **kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = RawReader(self._object)
@@ -152,7 +182,9 @@ class BufferedInputBase(io.BufferedIOBase):
 
     def read(self, size=-1):
         """Read up to size bytes from the object and return them."""
-        if size <= 0:
+        if size == 0:
+            return b''
+        elif size < 0:
             if len(self._buffer):
                 from_buf = self._read_from_buffer(len(self._buffer))
             else:
@@ -207,12 +239,10 @@ class BufferedInputBase(io.BufferedIOBase):
             #
             if self._line_terminator in self._buffer:
                 next_newline = self._buffer.index(self._line_terminator)
-                the_line.write(self._buffer[:next_newline + 1])
-                self._buffer = self._buffer[next_newline + 1:]
+                the_line.write(self._read_from_buffer(next_newline + 1))
                 break
             else:
-                the_line.write(self._buffer)
-                self._buffer = b''
+                the_line.write(self._read_from_buffer(len(self._buffer)))
                 self._fill_buffer(self._buffer_size)
         return the_line.getvalue()
 
@@ -250,7 +280,10 @@ class SeekableBufferedInputBase(BufferedInputBase):
 
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
                  line_terminator=BINARY_NEWLINE, **kwargs):
-        session = boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        session = kwargs.pop(
+            's3_session',
+            boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        )
         s3 = session.resource('s3', **kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = SeekableRawReader(self._object)
@@ -290,9 +323,10 @@ class SeekableBufferedInputBase(BufferedInputBase):
         else:
             new_position = self._content_length + offset
         new_position = _clamp(new_position, 0, self._content_length)
+        self._current_pos = new_position
+        self._raw_reader.seek(new_position)
+        logger.debug('new_position: %r', self._current_pos)
 
-        logger.debug('new_position: %r', new_position)
-        self._current_pos = self._raw_reader.position = new_position
         self._buffer = b""
         self._eof = self._current_pos == self._content_length
         return self._current_pos
@@ -311,12 +345,15 @@ class BufferedOutputBase(io.BufferedIOBase):
 
     Implements the io.BufferedIOBase interface of the standard library."""
 
-    def __init__(self, bucket, key, min_part_size=DEFAULT_MIN_PART_SIZE, **kwargs):
+    def __init__(self, bucket, key, min_part_size=DEFAULT_MIN_PART_SIZE, s3_upload=None, **kwargs):
         if min_part_size < MIN_MIN_PART_SIZE:
             logger.warning("S3 requires minimum part size >= 5MB; \
 multipart upload may fail")
 
-        session = boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        session = kwargs.pop(
+            's3_session',
+            boto3.Session(profile_name=kwargs.pop('profile_name', None))
+        )
         s3 = session.resource('s3', **kwargs)
 
         #
@@ -328,7 +365,7 @@ multipart upload may fail")
             raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket)
         self._object = s3.Object(bucket, key)
         self._min_part_size = min_part_size
-        self._mp = self._object.initiate_multipart_upload()
+        self._mp = self._object.initiate_multipart_upload(**(s3_upload or {}))
 
         self._buf = io.BytesIO()
         self._total_bytes = 0
@@ -433,3 +470,136 @@ multipart upload may fail")
             self.terminate()
         else:
             self.close()
+
+
+def iter_bucket(bucket_name, prefix='', accept_key=lambda key: True,
+                key_limit=None, workers=16, retries=3):
+    """
+    Iterate and download all S3 files under `bucket/prefix`, yielding out
+    `(key, key content)` 2-tuples (generator).
+
+    `accept_key` is a function that accepts a key name (unicode string) and
+    returns True/False, signalling whether the given key should be downloaded out or
+    not (default: accept all keys).
+
+    If `key_limit` is given, stop after yielding out that many results.
+
+    The keys are processed in parallel, using `workers` processes (default: 16),
+    to speed up downloads greatly. If multiprocessing is not available, thus
+    _MULTIPROCESSING is False, this parameter will be ignored.
+
+    Example::
+
+      >>> # get all JSON files under "mybucket/foo/"
+      >>> for key, content in iter_bucket(bucket_name, prefix='foo/', accept_key=lambda key: key.endswith('.json')):
+      ...     print key, len(content)
+
+      >>> # limit to 10k files, using 32 parallel workers (default is 16)
+      >>> for key, content in iter_bucket(bucket_name, key_limit=10000, workers=32):
+      ...     print key, len(content)
+    """
+    #
+    # If people insist on giving us bucket instances, silently extract the name
+    # before moving on.  Works for boto3 as well as boto.
+    #
+    try:
+        bucket_name = bucket_name.name
+    except AttributeError:
+        pass
+
+    total_size, key_no = 0, -1
+    key_iterator = _list_bucket(bucket_name, prefix=prefix, accept_key=accept_key)
+    download_key = functools.partial(_download_key, bucket_name=bucket_name, retries=retries)
+
+    with _create_process_pool(processes=workers) as pool:
+        result_iterator = pool.imap_unordered(download_key, key_iterator)
+        for key_no, (key, content) in enumerate(result_iterator):
+            if True or key_no % 1000 == 0:
+                logger.info(
+                    "yielding key #%i: %s, size %i (total %.1fMB)",
+                    key_no, key, len(content), total_size / 1024.0 ** 2
+                )
+            yield key, content
+            total_size += len(content)
+
+            if key_limit is not None and key_no + 1 >= key_limit:
+                # we were asked to output only a limited number of keys => we're done
+                break
+    logger.info("processed %i keys, total size %i" % (key_no + 1, total_size))
+
+
+def _list_bucket(bucket_name, prefix='', accept_key=lambda k: True):
+    client = boto3.client('s3')
+    ctoken = None
+
+    while True:
+        response = client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)
+        try:
+            content = response['Contents']
+        except KeyError:
+            pass
+        else:
+            for c in content:
+                key = c['Key']
+                if accept_key(key):
+                    yield key
+        ctoken = response.get('NextContinuationToken', None)
+        if not ctoken:
+            break
+
+
+def _download_key(key_name, bucket_name=None, retries=3):
+    if bucket_name is None:
+        raise ValueError('bucket_name may not be None')
+
+    #
+    # https://geekpete.com/blog/multithreading-boto3/
+    #
+    session = boto3.session.Session()
+    s3 = session.resource('s3')
+    bucket = s3.Bucket(bucket_name)
+
+    # Sometimes, https://github.com/boto/boto/issues/2409 can happen because of network issues on either side.
+    # Retry up to 3 times to ensure its not a transient issue.
+    for x in range(retries + 1):
+        try:
+            content_bytes = _download_fileobj(bucket, key_name)
+        except botocore.client.ClientError:
+            # Actually fail on last pass through the loop
+            if x == retries:
+                raise
+            # Otherwise, try again, as this might be a transient timeout
+            pass
+        else:
+            return key_name, content_bytes
+
+
+def _download_fileobj(bucket, key_name):
+    #
+    # This is a separate function only because it makes it easier to inject
+    # exceptions during tests.
+    #
+    buf = io.BytesIO()
+    bucket.download_fileobj(key_name, buf)
+    return buf.getvalue()
+
+
+class DummyPool(object):
+    """A class that mimics multiprocessing.pool.Pool for our purposes."""
+    def imap_unordered(self, function, items):
+        return six.moves.map(function, items)
+
+    def terminate(self):
+        pass
+
+
+@contextlib.contextmanager
+def _create_process_pool(processes=1):
+    if _MULTIPROCESSING and processes:
+        logger.info("creating pool with %i workers", processes)
+        pool = multiprocessing.pool.Pool(processes=processes)
+    else:
+        logger.info("creating dummy pool")
+        pool = DummyPool()
+    yield pool
+    pool.terminate()
